//------------------------------------------------------------------------------------------------------------------------------
// PROGRAMME ILLUSTRANT LES REGLES DE RASTERISATION
//
// LE PROGRAMME AFFICHE - UN ENSEMBLE DE LIGNES VERTICALES ET HORIZONTALES DESSINANT UNE GRILLE
//                      - UN CARRE ROUGE
//                      - UN POINT BLEU
// LES GEOMETRIES ET LES TRANSFORMATIONS SONT TELLES QUE
//                      - LE POINT EST CENTRÉ SUR LE CENTRE D'UN PIXEL
//                      - LES LIGNES PASSENT PAR LES CENTRES DES PIXELS, ET COMMENCENT ET S'ARRETENT PAR DES CENTRES DE PIXELS
//                      - LES CONTOURS DU CARRE SONT ALIGNÉS AVEC LES CONTOURS DES PIXELS
//
// LA VARIABLE DistanceAuCentreDuPixel (VOIR DANS LE CODE) PEUT ETRE MODIFIÉE POUR CONFIRMER L'EFFET DES REGLES DE RASTERISATION
// ELLE PERMET DE DECALER LES LIGNES PAR RAPPORT AU CENTRE DES PIXELS
//
//------------------------------------------------------------------------------------------------------------------------------

#include <iostream> // pour utiliser std::cout std::endl << >>

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include "main.h"
#include "LisEtCompileLesShaders.h"


#define TAILLE_EN_PIXEL 512

GLFWwindow* MaFenetre;

// VARIABLES PERMETTANT DE TRANSMETTRE DES CONSTANTES AUX SHADERS GLSL
GLuint CleDeMaCouleurFinale;
GLuint CleDeHauteur;
GLuint CleDeLargeur;

// VARIABLES PERMETTANT D'INITIALISER ET DE DESIGNER LES BUFFERS GRAPHIQUES
GLuint QuadVAO;
GLuint GrilleVAO;
GLuint PointVAO;

// VARIABLE DESIGNANT LE PROGRAMME GLSL
GLuint MonProgrammeGLSL;

// LARGEUR ET HAUTEUR ACTUALISÉES DE LA FENETRE GRAPHIQUE
GLuint Largeur, Hauteur;

int main( int argc, char *argv[])
{
	// INITIALISATION DE LA FENETRE ET DU CONTEXTE OPENGL
	GLFWwindow *MaFenetre = InitialiseUnContexteEtUneFenetreOpenGL();


	// INITIALISATION DU PROGRAMME DE RENDU GLSL
	MonProgrammeGLSL = LisEtCompileLesShaders( "../shader/TraitementDesSommets.glsl", "../shader/TraitementDesFragments.glsl");
	glUseProgram(MonProgrammeGLSL);

	std::cout << "MonProgrammeGLSL = " << MonProgrammeGLSL << std::endl;

	// ON VA CHERCHER DANS LE PROGRAMME GLSL LA REFERENCE VERS LES CONSTANTES STOCKANT LA TAILLE DE LA FENETRE GRAPHIQUE
	CleDeLargeur = glGetUniformLocation( MonProgrammeGLSL, "Largeur");
	std::cout << "CleDeLargeur = " << CleDeLargeur << std::endl;
	CleDeHauteur = glGetUniformLocation( MonProgrammeGLSL, "Hauteur");
	std::cout << "CleDeHauteur = " << CleDeHauteur << std::endl;

	Largeur = TAILLE_EN_PIXEL; Hauteur = TAILLE_EN_PIXEL;

	// ON VA CHERCHER DANS LE PROGRAMME GLSL LA REFERENCE VERS LA VARIABLE CONTENANT LA COULEUR
	GLuint CleDeMaCouleurFinale = glGetUniformLocation( MonProgrammeGLSL, "MaCouleurFinale");
	std::cout << "CleDeMaCouleurFinale = " << CleDeMaCouleurFinale << std::endl;

	// INITIALISATION DES GEOMETRIES: QUAD, GRILLE ET POINT
	GLuint QuadVBO = TransfertLeQuadSurLaCarteGraphique();
	std::cout << "QuadVBO = " << QuadVBO << std::endl;
	QuadVAO = EncapsuleLaGeometrieDansUnVao( QuadVBO);
	std::cout<<"QuadVAO = "<<QuadVAO<< std::endl;

	GLuint GrilleVBO = TransfertLaGrilleSurLaCarteGraphique();
	std::cout << "GrilleVBO = " << GrilleVBO << std::endl;
	GrilleVAO = EncapsuleLaGeometrieDansUnVao( GrilleVBO);
	std::cout<<"GrilleVAO = "<< GrilleVAO<< std::endl;

	GLuint PointVBO = TransfertLePointSurLaCarteGraphique();
	std::cout << "PointVBO = " << PointVBO << std::endl;
	PointVAO = EncapsuleLaGeometrieDansUnVao( PointVBO);
	std::cout<<"PointVAO = "<< PointVAO<< std::endl;

	// BOUCLE DE RENDU
	while (!glfwWindowShouldClose( MaFenetre)) {

		TraiteLesEvenementsClavier( MaFenetre);

		Rendu();

// EN CHOISISSANT "PollEvents" LE REDESSIN EST SYSTEMATIQUE
// EN CHOISISSANT "WaitEvents" LE REDESSIN N A LIEU QUE SI UN EVENEMENT EST EMIS
//		glfwPollEvents();
		glfwWaitEvents();
	}


	// SUPPRESSION DES BUFFERS ALLOUES SUR LA CARTE
	glDeleteBuffers(1, &QuadVBO);
	glDeleteVertexArrays(1, &QuadVAO);
	glDeleteBuffers(1, &PointVBO);
	glDeleteVertexArrays(1, &PointVAO);
	glDeleteBuffers(1, &GrilleVBO);
	glDeleteVertexArrays(1, &GrilleVAO);

	glfwTerminate();
}

void Rendu()
{
	static int nRendu = 0;

	std::cout << "je lance le rendu #"<< nRendu++ << "                            \r";
	std::cout.flush();

	glfwMakeContextCurrent(MaFenetre);

	// "EFFACAGE" DE LA FENETRE
	glClear( GL_COLOR_BUFFER_BIT );

	// ACTIVATION DU SHADER 
	glUseProgram(MonProgrammeGLSL);

	// DESSIN DE LA GRILLE EN NOIR PUIS DU CARRE EN ROUGE ET EN DERNIER DU POINT EN BLEU

	glBindVertexArray( GrilleVAO);
	// LE 1 INDIQUE QU'UNE SEULE COULEUR EST TRANSMISE VIA CETTE CLE/CE POINTEUR
	glUniform4fv( CleDeMaCouleurFinale, 1, glm::value_ptr(glm::vec4(0.,0.,0.,1.)));

	// TRANSMISSION DE LA TAILLE ACTUELLE DE LA FENETRE
	glUniform1i( CleDeLargeur, Largeur);
	glUniform1i( CleDeHauteur, Hauteur);

	// LES 260 SOMMETS DE LA GRILLE SONT PARCOURUS POUR PRODUIRE LES 65 SEGMENTS HORIZONTAUX ET LES 65 SEGMENTS VERTICAUX
	glDrawArrays( GL_LINES, 0, 260);

	glBindVertexArray( QuadVAO);
	glUniform4fv( CleDeMaCouleurFinale, 1, glm::value_ptr(glm::vec4(1.,0.,0.,1.)));
	// LES 4 SOMMETS DU RECTANGLE PERMETTENT DE DEFINIR 2 TRIANGLES COMPOSANT LE CARRE
	// LE PREMIER TRIANGLE EST CONSTITUE DES SOMMETS 0, 1 ET 2.
	// LE SECOND TRIANGLE DES SOMMETS 1, 2 ET 3
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4);

	glBindVertexArray( PointVAO);
	glUniform4fv( CleDeMaCouleurFinale, 1, glm::value_ptr(glm::vec4(0.,0.,1.,1.)));
	glDrawArrays( GL_POINTS, 0, 1); // POUR LA TAILLE DES POINTS VOIR gl_PointSize DANS LE VERTEX SHADER


	glBindVertexArray( 0);

	glfwSwapBuffers(MaFenetre);
}

GLuint TransfertLeQuadSurLaCarteGraphique()
{
	#define GRAND TAILLE_EN_PIXEL/4 + TAILLE_EN_PIXEL / 2.
	#define PETIT TAILLE_EN_PIXEL/4
	glm::vec3 MesSommetsQuad[]= { 
		glm::vec3( GRAND, PETIT, 0.0),
		glm::vec3( GRAND,  GRAND, 0.0),
		glm::vec3( PETIT,  PETIT, 0.0),
		glm::vec3( PETIT,  GRAND, 0.0)
	};

	GLuint vboID;
	glGenBuffers( 1, &vboID);
	glBindBuffer(GL_ARRAY_BUFFER, vboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MesSommetsQuad), MesSommetsQuad, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return vboID;
}

GLuint TransfertLaGrilleSurLaCarteGraphique()
{
	int x, y;
	int i = 0;

	// LA GRILLE COMPORTE 65 LIGNES VERTICALES (POUR Y VARIANT DE 0 A 512 INCLUS PAR INCREMENT DE 8)
	// ET 65 LIGNES HORIZONTALES. SOIT UN TOTAL DE 130 LIGNES ET DONC 260 SOMMETS
	glm::vec3 MesSommetsGrille[260];

	// les lignes passent exactement par le centre des pixels
	// LIGNES VERTICALES	
	for( x=0; x<=512; x+=8 ) {
		// ON PEUT MODIFIER A 0.7 OU -0.7 LA VARIABLE SUIVANTE POUR VOIR L'EFFET SUR L'IMAGE
		// POUR TOUTES LES VALEURS DE DistanceAuCentreDuPixel ENTRE -0.5 ET 0.5, LE DESSIN EST INCHANGE
		// => CELA ILLUSTRE QUE LE CENTRE MATHEMATIQUE DES PIXELS EST BIEN SITUE A 1/2 1/2
		float DistanceAuCentreDuPixel = 0.0;
		MesSommetsGrille[i++] = glm::vec3(x+0.5+DistanceAuCentreDuPixel,0.5,0.);
		MesSommetsGrille[i++] = glm::vec3(x+0.5+DistanceAuCentreDuPixel,512.5,0.);
	}

	// LIGNES HORIZONTALES
	for( y=0; y<=512; y+=8 ) {
		// ON PEUT MODIFIER A 0.7 OU -0.7 LA VARIABLE SUIVANTE POUR VOIR L'EFFET SUR L'IMAGE
		float DistanceAuCentreDuPixel = 0.0;
		MesSommetsGrille[i++] = glm::vec3(0.5,y+0.5+DistanceAuCentreDuPixel,0.);
		MesSommetsGrille[i++] = glm::vec3(512.5,y+0.5+DistanceAuCentreDuPixel,0.);
	}

	GLuint vboID;
	glGenBuffers( 1, &vboID);
	glBindBuffer(GL_ARRAY_BUFFER, vboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MesSommetsGrille), MesSommetsGrille, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	return vboID;
}

GLuint TransfertLePointSurLaCarteGraphique()
{
	// LE POINT EST CENTRE SUR LE CENTRE D'UN PIXEL
	glm::vec3 MonSommet[1] = {glm::vec3( 400.5, 256.5, 0.)};

	GLuint vboID;
	glGenBuffers( 1, &vboID);
	glBindBuffer(GL_ARRAY_BUFFER, vboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MonSommet), MonSommet, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return vboID;
}

GLuint EncapsuleLaGeometrieDansUnVao(GLuint CleDeMaGeometrie)
{
	GLuint vaoID;

	glGenVertexArrays( 1, &vaoID);
	glBindVertexArray(vaoID);

	glEnableVertexAttribArray(0); // ID de l'attribut a configurer

        glBindBuffer(GL_ARRAY_BUFFER, CleDeMaGeometrie); // ON ACTIVE LE VBO

       // On lui indique comment lire les donnees
	glVertexAttribPointer(
            0,                  // indice de l'attribut a configurer, correspond a la ligne "layout(location=0) in vec3 MonSommetEnEntree;" dans le vertex shader
            3,                  // nombre de composante par position (x, y, z)
            GL_FLOAT,           // type des composantes
            GL_FALSE,           // normalisation des composantes
            0,                  // decalage des composantes
            (void*)0            // offset des composantes
        );
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);

	return vaoID;
}

void UneFenetreChangeDeTaille( GLFWwindow* UneFenetre, int NouvelleLargeur, int NouvelleHauteur)
{
	Largeur = NouvelleLargeur; Hauteur = NouvelleHauteur;
// AFFICHAGE DE LA NOUVELLE TAILLE DE FENETRE
//	std::cout << "Nouvelle taille de fenetre: "<< Largeur << " x " << Hauteur << std::endl;

	Rendu();
}

GLFWwindow* InitialiseUnContexteEtUneFenetreOpenGL()
{
	
	// Initialisation de GLFW
	if( !glfwInit() )
	{
		std::cout << "Echec de l'initialisation de GLFW" << std::endl;
		exit(EXIT_FAILURE);
	}

//	glfwWindowHint(GLFW_SAMPLES, 4); // Anti Aliasing
	glfwWindowHint(GLFW_SAMPLES, 0); // AUCUN ANTI-ALIASING - TRES IMPORTANT POUR CE PROGRAMME DE TEST DE RASTERISATION
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // OpenGL 3.3
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);


	// Ouverture d'une fenêtre en 512x512
	// et creation d'un contexte OpenGL
	if( !(MaFenetre = glfwCreateWindow(TAILLE_EN_PIXEL, TAILLE_EN_PIXEL, "Polytech_INFO4", NULL,NULL)) )
	{
		std::cout << "Echec de l'ouverture de fenetre OpenGL" << std::endl;
		glfwTerminate();
		exit(EXIT_FAILURE);
	}

    	// Autorise GLFW a recevoir les appuis de touche
	glfwSetInputMode( MaFenetre, GLFW_STICKY_KEYS, GL_TRUE);

	// PROCEDURE APPELLEE QUAND LA FENETRE CHANGE DE TAILLE
	glfwSetFramebufferSizeCallback( MaFenetre, UneFenetreChangeDeTaille);

    // Initialisation de  GLEW
	glfwMakeContextCurrent(MaFenetre);
	glewExperimental = GL_TRUE; 
	if (glewInit() != GLEW_OK) {
		std::cout << "Echec de l'initialisation de GLEW" << std::endl;
		exit(EXIT_FAILURE);
	}

	glClearColor( 1., 1., 1., 0.);

	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);


	// VOIR http://www.glfw.org/docs/latest/quick.html#quick_render POUR LES EXPLICATIONS SUR LA LIGNE SUIVANTE
	glfwSwapInterval(1);

	// Affichage des donnees du contexte OpenGL
	const GLubyte* renderer = glGetString (GL_RENDERER);
	std::cout << "Carte Graphique : " << renderer << std::endl;

	const GLubyte* version = glGetString (GL_VERSION);
	std::cout << "Driver OpenGL : " << version << std::endl;

	const GLubyte* shading_version = glGetString (GL_SHADING_LANGUAGE_VERSION);
	std::cout << "Shading Language: " << shading_version << std::endl;

	return MaFenetre;
}

void TraiteLesEvenementsClavier(GLFWwindow *MaFenetre)
{
	// A TITRE D'EXEMPLE ON INDIQUE COMMENT DETECTER SI LA TOUCHE 'A' EST PRESSEE LORSQUE LA FENETRE EST ACTIVE
	if (glfwGetKey( MaFenetre, GLFW_KEY_A ) == GLFW_PRESS) {
		std::cout << "Touche A\n";
	}
}


